@@ -1,6 +1,6 @@
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  *   Mupen64plus-ui-console - core_interface.c                             *
- *   Mupen64Plus homepage: http://code.google.com/p/mupen64plus/           *
+ *   Mupen64Plus homepage: https://mupen64plus.org/                        *
  *   Copyright (C) 2009 Richard Goedeken                                   *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
@@ -23,6 +23,10 @@
  * library and pointers to the core functions
  */
 
+#if defined(__APPLE__)
+    #include <CoreFoundation/CoreFoundation.h>
+#endif
+
 #include <stdio.h>
 
 #include "core_interface.h"
@@ -73,6 +77,11 @@ ptr_ConfigGetParamFloat    ConfigGetParamFloat = NULL;
 ptr_ConfigGetParamBool     ConfigGetParamBool = NULL;
 ptr_ConfigGetParamString   ConfigGetParamString = NULL;
 
+ptr_ConfigExternalOpen         ConfigExternalOpen = NULL;
+ptr_ConfigExternalClose        ConfigExternalClose = NULL;
+ptr_ConfigExternalGetParameter ConfigExternalGetParameter = NULL;
+ptr_ConfigHasUnsavedChanges    ConfigHasUnsavedChanges = NULL;
+
 ptr_ConfigGetSharedDataFilepath ConfigGetSharedDataFilepath = NULL;
 ptr_ConfigGetUserConfigPath     ConfigGetUserConfigPath = NULL;
 ptr_ConfigGetUserDataPath       ConfigGetUserDataPath = NULL;
@@ -103,6 +112,9 @@ ptr_DebugGetCPUDataPtr     DebugGetCPUDataPtr = NULL;
 ptr_DebugBreakpointLookup  DebugBreakpointLookup = NULL;
 ptr_DebugBreakpointCommand DebugBreakpointCommand = NULL;
 
+ptr_DebugBreakpointTriggeredBy DebugBreakpointTriggeredBy = NULL;
+ptr_DebugVirtualToPhysical     DebugVirtualToPhysical = NULL;
+
 /* global variables */
 m64p_dynlib_handle CoreHandle = NULL;
 
@@ -126,6 +138,24 @@ m64p_error AttachCoreLib(const char *CoreLibFilepath)
     {
         rval = osal_dynlib_open(&CoreHandle, COREDIR OSAL_DEFAULT_DYNLIB_FILENAME);
     }
+#endif
+    /* for MacOS, look for the library in the Frameworks folder of the app bundle */
+#if defined(__APPLE__)
+    CFBundleRef mainBundle = CFBundleGetMainBundle();
+    if (mainBundle != NULL)
+    {
+        CFURLRef frameworksURL = CFBundleCopyPrivateFrameworksURL(mainBundle);
+        if (frameworksURL != NULL)
+        {
+            char libPath[1024 + 32];
+            if (CFURLGetFileSystemRepresentation(frameworksURL, TRUE, (uint8_t *) libPath, 1024))
+            {
+                strcat(libPath, "/" OSAL_DEFAULT_DYNLIB_FILENAME);
+                rval = osal_dynlib_open(&CoreHandle, libPath);
+            }
+            CFRelease(frameworksURL);
+        }
+    }
 #endif
     /* then try just the filename of the shared library, to let dlopen() look through the system lib dirs */
     if (rval != M64ERR_SUCCESS || CoreHandle == NULL)
@@ -239,6 +269,11 @@ m64p_error AttachCoreLib(const char *CoreLibFilepath)
     ConfigGetParamBool = (ptr_ConfigGetParamBool) osal_dynlib_getproc(CoreHandle, "ConfigGetParamBool");
     ConfigGetParamString = (ptr_ConfigGetParamString) osal_dynlib_getproc(CoreHandle, "ConfigGetParamString");
 
+    ConfigExternalOpen = (ptr_ConfigExternalOpen) osal_dynlib_getproc(CoreHandle, "ConfigExternalOpen");
+    ConfigExternalClose = (ptr_ConfigExternalClose) osal_dynlib_getproc(CoreHandle, "ConfigExternalClose");
+    ConfigExternalGetParameter = (ptr_ConfigExternalGetParameter) osal_dynlib_getproc(CoreHandle, "ConfigExternalGetParameter");
+    ConfigHasUnsavedChanges = (ptr_ConfigHasUnsavedChanges) osal_dynlib_getproc(CoreHandle, "ConfigHasUnsavedChanges");
+
     ConfigGetSharedDataFilepath = (ptr_ConfigGetSharedDataFilepath) osal_dynlib_getproc(CoreHandle, "ConfigGetSharedDataFilepath");
     ConfigGetUserConfigPath = (ptr_ConfigGetUserConfigPath) osal_dynlib_getproc(CoreHandle, "ConfigGetUserConfigPath");
     ConfigGetUserDataPath = (ptr_ConfigGetUserDataPath) osal_dynlib_getproc(CoreHandle, "ConfigGetUserDataPath");
@@ -260,15 +295,18 @@ m64p_error AttachCoreLib(const char *CoreLibFilepath)
     DebugMemRead16 = (ptr_DebugMemRead16) osal_dynlib_getproc(CoreHandle, "DebugMemRead16");
     DebugMemRead8 = (ptr_DebugMemRead8) osal_dynlib_getproc(CoreHandle, "DebugMemRead8");
 
-    DebugMemWrite64 = (ptr_DebugMemWrite64) osal_dynlib_getproc(CoreHandle, "DebugMemRead64");
-    DebugMemWrite32 = (ptr_DebugMemWrite32) osal_dynlib_getproc(CoreHandle, "DebugMemRead32");
-    DebugMemWrite16 = (ptr_DebugMemWrite16) osal_dynlib_getproc(CoreHandle, "DebugMemRead16");
-    DebugMemWrite8 = (ptr_DebugMemWrite8) osal_dynlib_getproc(CoreHandle, "DebugMemRead8");
+    DebugMemWrite64 = (ptr_DebugMemWrite64) osal_dynlib_getproc(CoreHandle, "DebugMemWrite64");
+    DebugMemWrite32 = (ptr_DebugMemWrite32) osal_dynlib_getproc(CoreHandle, "DebugMemWrite32");
+    DebugMemWrite16 = (ptr_DebugMemWrite16) osal_dynlib_getproc(CoreHandle, "DebugMemWrite16");
+    DebugMemWrite8 = (ptr_DebugMemWrite8) osal_dynlib_getproc(CoreHandle, "DebugMemWrite8");
 
     DebugGetCPUDataPtr = (ptr_DebugGetCPUDataPtr) osal_dynlib_getproc(CoreHandle, "DebugGetCPUDataPtr");
     DebugBreakpointLookup = (ptr_DebugBreakpointLookup) osal_dynlib_getproc(CoreHandle, "DebugBreakpointLookup");
     DebugBreakpointCommand = (ptr_DebugBreakpointCommand) osal_dynlib_getproc(CoreHandle, "DebugBreakpointCommand");
 
+    DebugBreakpointTriggeredBy = (ptr_DebugBreakpointTriggeredBy) osal_dynlib_getproc(CoreHandle, "DebugBreakpointTriggeredBy");
+    DebugVirtualToPhysical = (ptr_DebugVirtualToPhysical) osal_dynlib_getproc(CoreHandle, "DebugVirtualToPhysical");
+
     return M64ERR_SUCCESS;
 }
 
@@ -304,6 +342,11 @@ m64p_error DetachCoreLib(void)
     ConfigGetParamBool = NULL;
     ConfigGetParamString = NULL;
 
+    ConfigExternalOpen = NULL;
+    ConfigExternalClose = NULL;
+    ConfigExternalGetParameter = NULL;
+    ConfigHasUnsavedChanges = NULL;
+
     ConfigGetSharedDataFilepath = NULL;
     ConfigGetUserDataPath = NULL;
     ConfigGetUserCachePath = NULL;
@@ -332,6 +375,9 @@ m64p_error DetachCoreLib(void)
     DebugBreakpointLookup = NULL;
     DebugBreakpointCommand = NULL;
 
+    DebugBreakpointTriggeredBy = NULL;
+    DebugVirtualToPhysical = NULL;
+
     /* detach the shared library */
     osal_dynlib_close(CoreHandle);
     CoreHandle = NULL;
