@@ -1,7 +1,7 @@
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  *   Mupen64plus-ui-console - main.c                                       *
- *   Mupen64Plus homepage: http://code.google.com/p/mupen64plus/           *
- *   Copyright (C) 2007-2010 Richard42                                     *
+ *   Mupen64Plus homepage: https://mupen64plus.org/                        *
+ *   Copyright (C) 2007-2018 Richard42                                     *
  *   Copyright (C) 2008 Ebenblues Nmn Okaygo Tillin9                       *
  *   Copyright (C) 2002 Hacktarux                                          *
  *                                                                         *
@@ -22,24 +22,35 @@
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
 /* This is the main application entry point for the console-only front-end
- * for Mupen64Plus v2.0. 
+ * for Mupen64Plus v2.0.
  */
- 
+
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "SDL_main.h"
+#include <SDL.h>
+#include <SDL_main.h>
+#include <SDL_thread.h>
+
 #include "cheat.h"
 #include "compare_core.h"
 #include "core_interface.h"
+#include "debugger.h"
 #include "m64p_types.h"
 #include "main.h"
 #include "osal_preproc.h"
+#include "osal_files.h"
 #include "plugin.h"
 #include "version.h"
 
+#ifdef VIDEXT_HEADER
+#define xstr(s) str(s)
+#define str(s) #s
+#include xstr(VIDEXT_HEADER)
+#endif
+
 /* Version number for UI-Console config section parameters */
 #define CONFIG_PARAM_VERSION     1.00
 
@@ -50,6 +61,8 @@ int    g_Verbose = 0;
 static m64p_handle l_ConfigCore = NULL;
 static m64p_handle l_ConfigVideo = NULL;
 static m64p_handle l_ConfigUI = NULL;
+static m64p_handle l_ConfigTransferPak = NULL;
+static m64p_handle l_Config64DD = NULL;
 
 static const char *l_CoreLibPath = NULL;
 static const char *l_ConfigDirPath = NULL;
@@ -66,6 +79,7 @@ static int  *l_TestShotList = NULL;      // list of screenshots to take for regr
 static int   l_TestShotIdx = 0;          // index of next screenshot frame in list
 static int   l_SaveOptions = 1;          // save command-line options in configuration file (enabled by default)
 static int   l_CoreCompareMode = 0;      // 0 = disable, 1 = send, 2 = receive
+static int   l_LaunchDebugger = 0;
 
 static eCheatMode l_CheatMode = CHEAT_DISABLE;
 static char      *l_CheatNumList = NULL;
@@ -126,15 +140,6 @@ void DebugCallback(void *Context, int level, const char *message)
 
 static void FrameCallback(unsigned int FrameIndex)
 {
-    // load savestate at first frame if needed
-    if (l_SaveStatePath != NULL && FrameIndex == 0)
-    {
-        if ((*CoreDoCommand)(M64CMD_STATE_LOAD, 0, (void *) l_SaveStatePath) != M64ERR_SUCCESS)
-        {
-            DebugMessage(M64MSG_WARNING, "couldn't load state, rom will run normally.");
-        }
-    }
-
     // take a screenshot if we need to
     if (l_TestShotList != NULL)
     {
@@ -154,6 +159,69 @@ static void FrameCallback(unsigned int FrameIndex)
     }
 }
 
+
+static char *formatstr(const char *fmt, ...)
+{
+	int size = 128, ret;
+	char *str = (char *)malloc(size), *newstr;
+	va_list args;
+
+	/* There are two implementations of vsnprintf we have to deal with:
+	 * C99 version: Returns the number of characters which would have been written
+	 *              if the buffer had been large enough, and -1 on failure.
+	 * Windows version: Returns the number of characters actually written,
+	 *                  and -1 on failure or truncation.
+	 * NOTE: An implementation equivalent to the Windows one appears in glibc <2.1.
+	 */
+	while (str != NULL)
+	{
+		va_start(args, fmt);
+		ret = vsnprintf(str, size, fmt, args);
+		va_end(args);
+
+		// Successful result?
+		if (ret >= 0 && ret < size)
+			return str;
+
+		// Increment the capacity of the buffer
+		if (ret >= size)
+			size = ret + 1; // C99 version: We got the needed buffer size
+		else
+			size *= 2; // Windows version: Keep guessing
+
+		newstr = (char *)realloc(str, size);
+		if (newstr == NULL)
+			free(str);
+		str = newstr;
+	}
+
+	return NULL;
+}
+
+static int is_path_separator(char c)
+{
+    return strchr(OSAL_DIR_SEPARATORS, c) != NULL;
+}
+
+char* combinepath(const char* first, const char *second)
+{
+    size_t len_first, off_second = 0;
+
+    if (first == NULL || second == NULL)
+        return NULL;
+
+    len_first = strlen(first);
+
+    while (is_path_separator(first[len_first-1]))
+        len_first--;
+
+    while (is_path_separator(second[off_second]))
+        off_second++;
+
+    return formatstr("%.*s%c%s", (int) len_first, first, OSAL_DIR_SEPARATORS[0], second + off_second);
+}
+
+
 /*********************************************************************************************************
  *  Configuration handling
  */
@@ -163,6 +231,7 @@ static m64p_error OpenConfigurationHandles(void)
     float fConfigParamsVersion;
     int bSaveConfig = 0;
     m64p_error rval;
+    unsigned int i;
 
     /* Open Configuration sections for core library and console User Interface */
     rval = (*ConfigOpenSection)("Core", &l_ConfigCore);
@@ -179,6 +248,20 @@ static m64p_error OpenConfigurationHandles(void)
         return rval;
     }
 
+    rval = (*ConfigOpenSection)("Transferpak", &l_ConfigTransferPak);
+    if (rval != M64ERR_SUCCESS)
+    {
+        DebugMessage(M64MSG_ERROR, "failed to open 'Transferpak' configuration section");
+        return rval;
+    }
+
+    rval = (*ConfigOpenSection)("64DD", &l_Config64DD);
+    if (rval != M64ERR_SUCCESS)
+    {
+        DebugMessage(M64MSG_ERROR, "failed to open '64DD' configuration section");
+        return rval;
+    }
+
     rval = (*ConfigOpenSection)("UI-Console", &l_ConfigUI);
     if (rval != M64ERR_SUCCESS)
     {
@@ -217,8 +300,28 @@ static m64p_error OpenConfigurationHandles(void)
     (*ConfigSetDefaultString)(l_ConfigUI, "InputPlugin", "mupen64plus-input-sdl" OSAL_DLL_EXTENSION, "Filename of input plugin");
     (*ConfigSetDefaultString)(l_ConfigUI, "RspPlugin", "mupen64plus-rsp-hle" OSAL_DLL_EXTENSION, "Filename of RSP plugin");
 
-    if (bSaveConfig && ConfigSaveSection != NULL) /* ConfigSaveSection was added in Config API v2.1.0 */
+    for(i = 1; i < 5; ++i) {
+        char key[64];
+        char desc[2048];
+#define SET_DEFAULT_STRING(key_fmt, default_value, desc_fmt) \
+        do { \
+            snprintf(key, sizeof(key), key_fmt, i); \
+            snprintf(desc, sizeof(desc), desc_fmt, i); \
+            (*ConfigSetDefaultString)(l_ConfigTransferPak, key, default_value, desc); \
+        } while(0)
+
+        SET_DEFAULT_STRING("GB-rom-%u", "", "Filename of the GB ROM to load into transferpak %u");
+        SET_DEFAULT_STRING("GB-ram-%u", "", "Filename of the GB RAM to load into transferpak %u");
+#undef SET_DEFAULT_STRING
+    }
+
+    (*ConfigSetDefaultString)(l_Config64DD, "IPL-ROM", "", "Filename of the 64DD IPL ROM");
+    (*ConfigSetDefaultString)(l_Config64DD, "Disk", "", "Filename of the disk to load into Disk Drive");
+
+    if (bSaveConfig && l_SaveOptions && ConfigSaveSection != NULL) { /* ConfigSaveSection was added in Config API v2.1.0 */
         (*ConfigSaveSection)("UI-Console");
+        (*ConfigSaveSection)("Transferpak");
+    }
 
     return M64ERR_SUCCESS;
 }
@@ -241,7 +344,10 @@ static m64p_error SaveConfigurationOptions(void)
     if (g_RspPlugin != NULL)
         (*ConfigSetParameter)(l_ConfigUI, "RspPlugin", M64TYPE_STRING, g_RspPlugin);
 
-    return (*ConfigSaveFile)();
+    if ((*ConfigHasUnsavedChanges)(NULL))
+        return (*ConfigSaveFile)();
+    else
+        return M64ERR_SUCCESS;
 }
 
 /*********************************************************************************************************
@@ -263,6 +369,7 @@ static void printUsage(const char *progname)
            "    --corelib (filepath)   : use core library (filepath) (can be only filename or full path)\n"
            "    --configdir (dir)      : force configation directory to (dir); should contain mupen64plus.cfg\n"
            "    --datadir (dir)        : search for shared data files (.ini files, languages, etc) in (dir)\n"
+           "    --debug                : launch console-based debugger (requires core lib built for debugging)\n"
            "    --plugindir (dir)      : search for plugins in (dir)\n"
            "    --sshotdir (dir)       : set screenshot directory to (dir)\n"
            "    --gfx (plugin-spec)    : use gfx plugin given by (plugin-spec)\n"
@@ -273,6 +380,10 @@ static void printUsage(const char *progname)
            "    --savestate (filepath) : savestate loaded at startup\n"
            "    --testshots (list)     : take screenshots at frames given in comma-separated (list), then quit\n"
            "    --set (param-spec)     : set a configuration variable, format: ParamSection[ParamName]=Value\n"
+           "    --gb-rom-{1,2,3,4}     : define GB cart rom to load inside transferpak {1,2,3,4}\n"
+           "    --gb-ram-{1,2,3,4}     : define GB cart ram to load inside transferpak {1,2,3,4}\n"
+           "    --dd-ipl-rom           : define 64DD IPL rom\n"
+           "    --dd-disk              : define disk to load into the disk drive\n"
            "    --core-compare-send    : use the Core Comparison debugging feature, in data sending mode\n"
            "    --core-compare-recv    : use the Core Comparison debugging feature, in data receiving mode\n"
            "    --nosaveoptions        : do not save the given command-line options in configuration file\n"
@@ -413,7 +524,8 @@ static int ParseCommandLineInitial(int argc, const char **argv)
 {
     int i;
 
-    /* look through commandline options */
+    /* First phase of command line parsing: read parameters that affect the
+       core and the ui-console behavior. */
     for (i = 1; i < argc; i++)
     {
         int ArgsLeft = argc - i - 1;
@@ -438,16 +550,21 @@ static int ParseCommandLineInitial(int argc, const char **argv)
             printUsage(argv[0]);
             return 1;
         }
+        else if (strcmp(argv[i], "--nosaveoptions") == 0)
+        {
+            l_SaveOptions = 0;
+        }
     }
 
     return 0;
 }
 
-static m64p_error ParseCommandLineFinal(int argc, const char **argv)
+static m64p_error ParseCommandLineMain(int argc, const char **argv)
 {
     int i;
 
-    /* parse commandline options */
+    /* Second phase of command-line parsing: read all remaining parameters
+       except for those that set plugin options. */
     for (i = 1; i < argc; i++)
     {
         int ArgsLeft = argc - i - 1;
@@ -484,9 +601,13 @@ static m64p_error ParseCommandLineFinal(int argc, const char **argv)
         }
         else if ((strcmp(argv[i], "--corelib") == 0 || strcmp(argv[i], "--configdir") == 0 ||
                   strcmp(argv[i], "--datadir") == 0) && ArgsLeft >= 1)
-        {   /* these are handled in ParseCommandLineInitial */
+        {   /* already handled in ParseCommandLineInitial (skip the value) */
             i++;
         }
+        else if (strcmp(argv[i], "--nosaveoptions") == 0)
+        {   /* already handled in ParseCommandLineInitial (no value to skip) */
+            ;
+        }
         else if (strcmp(argv[i], "--resolution") == 0 && ArgsLeft >= 1)
         {
             const char *res = argv[i+1];
@@ -571,10 +692,13 @@ static m64p_error ParseCommandLineFinal(int argc, const char **argv)
         }
         else if (strcmp(argv[i], "--set") == 0 && ArgsLeft >= 1)
         {
-            if (SetConfigParameter(argv[i+1]) != 0)
-                return M64ERR_INPUT_INVALID;
+            /* skip this: it will be handled in ParseCommandLinePlugin */
             i++;
         }
+        else if (strcmp(argv[i], "--debug") == 0)
+        {
+            l_LaunchDebugger = 1;
+        }
         else if (strcmp(argv[i], "--core-compare-send") == 0)
         {
             l_CoreCompareMode = 1;
@@ -583,9 +707,30 @@ static m64p_error ParseCommandLineFinal(int argc, const char **argv)
         {
             l_CoreCompareMode = 2;
         }
-        else if (strcmp(argv[i], "--nosaveoptions") == 0)
+#define PARSE_GB_CART_PARAM(param, key) \
+        else if (strcmp(argv[i], param) == 0) \
+        { \
+            ConfigSetParameter(l_ConfigTransferPak, key, M64TYPE_STRING, argv[i+1]); \
+            i++; \
+        }
+        PARSE_GB_CART_PARAM("--gb-rom-1", "GB-rom-1")
+        PARSE_GB_CART_PARAM("--gb-ram-1", "GB-ram-1")
+        PARSE_GB_CART_PARAM("--gb-rom-2", "GB-rom-2")
+        PARSE_GB_CART_PARAM("--gb-ram-2", "GB-ram-2")
+        PARSE_GB_CART_PARAM("--gb-rom-3", "GB-rom-3")
+        PARSE_GB_CART_PARAM("--gb-ram-3", "GB-ram-3")
+        PARSE_GB_CART_PARAM("--gb-rom-4", "GB-rom-4")
+        PARSE_GB_CART_PARAM("--gb-ram-4", "GB-ram-4")
+#undef PARSE_GB_CART_PARAM
+        else if (strcmp(argv[i], "--dd-ipl-rom") == 0)
         {
-            l_SaveOptions = 0;
+            ConfigSetParameter(l_Config64DD, "IPL-ROM", M64TYPE_STRING, argv[i+1]);
+            i++;
+        }
+        else if (strcmp(argv[i], "--dd-disk") == 0)
+        {
+            ConfigSetParameter(l_Config64DD, "Disk", M64TYPE_STRING, argv[i+1]);
+            i++;
         }
         else if (ArgsLeft == 0)
         {
@@ -609,6 +754,115 @@ static m64p_error ParseCommandLineFinal(int argc, const char **argv)
     return M64ERR_INPUT_INVALID;
 }
 
+static m64p_error ParseCommandLinePlugin(int argc, const char **argv)
+{
+    int i;
+
+    /* Third phase of command-line parsing: read all plugin parameters. */
+    for (i = 1; i < argc; i++)
+    {
+        int ArgsLeft = argc - i - 1;
+        if (strcmp(argv[i], "--set") == 0 && ArgsLeft >= 1)
+        {
+            if (SetConfigParameter(argv[i+1]) != 0)
+                return M64ERR_INPUT_INVALID;
+            i++;
+        }
+    }
+    return M64ERR_SUCCESS;
+}
+
+static char* media_loader_get_filename(void* cb_data, m64p_handle section_handle, const char* section, const char* key)
+{
+#define MUPEN64PLUS_CFG_NAME "mupen64plus.cfg"
+    m64p_handle core_config;
+    char value[4096];
+    const char* configdir = NULL;
+    char* cfgfilepath = NULL;
+
+    /* reset filename */
+    char* mem_filename = NULL;
+
+    /* XXX: use external config API to force reload of file content */
+    configdir = ConfigGetUserConfigPath();
+    if (configdir == NULL) {
+        DebugMessage(M64MSG_ERROR, "Can't get user config path !");
+        return NULL;
+    }
+
+    cfgfilepath = combinepath(configdir, MUPEN64PLUS_CFG_NAME);
+    if (cfgfilepath == NULL) {
+        DebugMessage(M64MSG_ERROR, "Can't get config file path: %s + %s!", configdir, MUPEN64PLUS_CFG_NAME);
+        return NULL;
+    }
+
+    if (ConfigExternalOpen(cfgfilepath, &core_config) != M64ERR_SUCCESS) {
+        DebugMessage(M64MSG_ERROR, "Can't open config file %s!", cfgfilepath);
+        goto release_cfgfilepath;
+    }
+
+    if (ConfigExternalGetParameter(core_config, section, key, value, sizeof(value)) != M64ERR_SUCCESS) {
+        DebugMessage(M64MSG_ERROR, "Can't get parameter %s", key);
+        goto close_config;
+    }
+
+    size_t len = strlen(value);
+    if (len < 2 || value[0] != '"' || value[len-1] != '"') {
+        DebugMessage(M64MSG_ERROR, "Invalid string format %s", value);
+        goto close_config;
+    }
+
+    value[len-1] = '\0';
+    mem_filename = strdup(value + 1);
+
+    ConfigSetParameter(section_handle, key, M64TYPE_STRING, mem_filename);
+
+close_config:
+    ConfigExternalClose(core_config);
+release_cfgfilepath:
+    free(cfgfilepath);
+    return mem_filename;
+}
+
+
+static char* media_loader_get_gb_cart_mem_file(void* cb_data, const char* mem, int control_id)
+{
+    char key[64];
+
+    snprintf(key, sizeof(key), "GB-%s-%u", mem, control_id + 1);
+    return media_loader_get_filename(cb_data, l_ConfigTransferPak, "Transferpak", key);
+}
+
+static char* media_loader_get_gb_cart_rom(void* cb_data, int control_id)
+{
+    return media_loader_get_gb_cart_mem_file(cb_data, "rom", control_id);
+}
+
+static char* media_loader_get_gb_cart_ram(void* cb_data, int control_id)
+{
+    return media_loader_get_gb_cart_mem_file(cb_data, "ram", control_id);
+}
+
+static char* media_loader_get_dd_rom(void* cb_data)
+{
+    return media_loader_get_filename(cb_data, l_Config64DD, "64DD", "IPL-ROM");
+}
+
+static char* media_loader_get_dd_disk(void* cb_data)
+{
+    return media_loader_get_filename(cb_data, l_Config64DD, "64DD", "Disk");
+}
+
+static m64p_media_loader l_media_loader =
+{
+    NULL,
+    media_loader_get_gb_cart_rom,
+    media_loader_get_gb_cart_ram,
+    media_loader_get_dd_rom,
+    media_loader_get_dd_disk
+};
+
+
 /*********************************************************************************************************
 * main function
 */
@@ -639,7 +893,7 @@ int main(int argc, char *argv[])
     printf("| |\\/| | | | | '_ \\ / _ \\ '_ \\| '_ \\| || |_| |_) | | | | / __|  \n");
     printf("| |  | | |_| | |_) |  __/ | | | (_) |__   _|  __/| | |_| \\__ \\  \n");
     printf("|_|  |_|\\__,_| .__/ \\___|_| |_|\\___/   |_| |_|   |_|\\__,_|___/  \n");
-    printf("             |_|         http://code.google.com/p/mupen64plus/  \n");
+    printf("             |_|         https://mupen64plus.org/               \n");
     printf("%s Version %i.%i.%i\n\n", CONSOLE_UI_NAME, VERSION_PRINTF_SPLIT(CONSOLE_UI_VERSION));
 
     /* bootstrap some special parameters from the command line */
@@ -659,6 +913,16 @@ int main(int argc, char *argv[])
         return 3;
     }
 
+#ifdef VIDEXT_HEADER
+    rval = CoreOverrideVidExt(&vidExtFunctions);
+    if (rval != M64ERR_SUCCESS)
+    {
+        DebugMessage(M64MSG_ERROR, "couldn't start VidExt library.");
+        DetachCoreLib();
+        return 14;
+    }
+#endif
+
     /* Open configuration sections */
     rval = OpenConfigurationHandles();
     if (rval != M64ERR_SUCCESS)
@@ -668,8 +932,8 @@ int main(int argc, char *argv[])
         return 4;
     }
 
-    /* parse command-line options */
-    rval = ParseCommandLineFinal(argc, (const char **) argv);
+    /* parse non-plugin command-line options */
+    rval = ParseCommandLineMain(argc, (const char **) argv);
     if (rval != M64ERR_SUCCESS)
     {
         (*CoreShutdown)();
@@ -677,7 +941,7 @@ int main(int argc, char *argv[])
         return 5;
     }
 
-    /* Handle the core comparison feature */
+    /* Ensure that the core supports comparison feature if necessary */
     if (l_CoreCompareMode != 0 && !(g_CoreCapabilities & M64CAPS_CORE_COMPARE))
     {
         DebugMessage(M64MSG_ERROR, "can't use --core-compare feature with this Mupen64Plus core library.");
@@ -685,6 +949,14 @@ int main(int argc, char *argv[])
         return 6;
     }
     compare_core_init(l_CoreCompareMode);
+    
+    /* Ensure that the core supports the debugger if necessary */
+    if (l_LaunchDebugger && !(g_CoreCapabilities & M64CAPS_DEBUGGER))
+    {
+        DebugMessage(M64MSG_ERROR, "can't use --debug feature with this Mupen64Plus core library.");
+        DetachCoreLib();
+        return 6;
+    }
 
     /* save the given command-line options in configuration file if requested */
     if (l_SaveOptions)
@@ -756,6 +1028,17 @@ int main(int argc, char *argv[])
         return 12;
     }
 
+    /* Parse and set plugin options. Doing this after loading the plugins
+       allows the plugins to set up their own defaults first. */
+    rval = ParseCommandLinePlugin(argc, (const char **) argv);
+    if (rval != M64ERR_SUCCESS)
+    {
+        (*CoreDoCommand)(M64CMD_ROM_CLOSE, 0, NULL);
+        (*CoreShutdown)();
+        DetachCoreLib();
+        return 5;
+    }
+
     /* attach plugins to core */
     for (i = 0; i < 4; i++)
     {
@@ -769,15 +1052,64 @@ int main(int argc, char *argv[])
         }
     }
 
-    /* set up Frame Callback if --testshots or --savestate is enabled */
-    if (l_TestShotList != NULL || l_SaveStatePath != NULL)
+    /* set up Frame Callback if --testshots is enabled */
+    if (l_TestShotList != NULL)
     {
         if ((*CoreDoCommand)(M64CMD_SET_FRAME_CALLBACK, 0, FrameCallback) != M64ERR_SUCCESS)
         {
-            DebugMessage(M64MSG_WARNING, "couldn't set frame callback, --testshots and/or --savestate will not work.");
+            DebugMessage(M64MSG_WARNING, "couldn't set frame callback, --testshots will not work.");
         }
     }
 
+    /* set gb cart loader */
+    if ((*CoreDoCommand)(M64CMD_SET_MEDIA_LOADER, sizeof(l_media_loader), &l_media_loader) != M64ERR_SUCCESS)
+    {
+        DebugMessage(M64MSG_WARNING, "Couldn't set media loader, transferpak and GB carts will not work.");
+    }
+
+    /* load savestate at startup */
+    if (l_SaveStatePath != NULL)
+    {
+        if ((*CoreDoCommand)(M64CMD_STATE_LOAD, 0, (void *) l_SaveStatePath) != M64ERR_SUCCESS)
+        {
+            DebugMessage(M64MSG_WARNING, "couldn't load state, rom will run normally.");
+        }
+    }
+
+    /* Setup debugger */
+    if (l_LaunchDebugger)
+    {
+        if (debugger_setup_callbacks())
+        {
+            DebugMessage(M64MSG_ERROR, "couldn't setup debugger callbacks.");
+            (*CoreDoCommand)(M64CMD_ROM_CLOSE, 0, NULL);
+            (*CoreShutdown)();
+            DetachCoreLib();
+            return 14;
+        }
+        /* Set Core config parameter to enable debugger */
+        int bEnableDebugger = 1;
+        (*ConfigSetParameter)(l_ConfigCore, "EnableDebugger", M64TYPE_BOOL, &bEnableDebugger);
+        /* Fork the debugger input thread. */
+#if SDL_VERSION_ATLEAST(2,0,0)
+        SDL_CreateThread(debugger_loop, "DebugLoop", NULL);
+#else
+        SDL_CreateThread(debugger_loop, NULL);
+#endif
+    }
+    else
+    {
+        /* Set Core config parameter to disable debugger */
+        int bEnableDebugger = 0;
+        (*ConfigSetParameter)(l_ConfigCore, "EnableDebugger", M64TYPE_BOOL, &bEnableDebugger);
+    }
+
+    /* Save the configuration file again, if necessary, to capture updated
+       parameters from plugins. This is the last opportunity to save changes
+       before the relatively long-running game. */
+    if (l_SaveOptions && (*ConfigHasUnsavedChanges)(NULL))
+        (*ConfigSaveFile)();
+
     /* run the game */
     (*CoreDoCommand)(M64CMD_EXECUTE, 0, NULL);
 
@@ -790,8 +1122,8 @@ int main(int argc, char *argv[])
     (*CoreDoCommand)(M64CMD_ROM_CLOSE, 0, NULL);
 
     /* save the configuration file again if --nosaveoptions was not specified, to keep any updated parameters from the core/plugins */
-    if (l_SaveOptions)
-        SaveConfigurationOptions();
+    if (l_SaveOptions && (*ConfigHasUnsavedChanges)(NULL))
+        (*ConfigSaveFile)();
 
     /* Shut down and release the Core library */
     (*CoreShutdown)();
